
'''
Copyright (c) 2021 yu9824

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.base import is_classifier, is_regressor
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_array, check_X_y, check_random_state
from sklearn.utils.validation import check_is_fitted
from sklearn.utils import Bunch
from sklearn.model_selection import train_test_split, cross_val_predict, cross_val_score
from sklearn.model_selection import check_cv
from sklearn.model_selection._validation import _fit_and_score, _aggregate_score_dicts, _score
from sklearn.metrics import check_scoring
from sklearn.metrics._scorer import _check_multimetric_scoring
from sklearn.ensemble import RandomForestRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from sklearn.linear_model import Ridge, Lasso, LinearRegression
from sklearn.tree import DecisionTreeRegressor
from sklearn.base import clone
from sklearn.svm import SVR
from sklearn.cross_decomposition import PLSRegression
from sklearn.inspection import permutation_importance
from ngboost import NGBRegressor
from lightgbm import LGBMRegressor
import sys

from yikit.feature_selection import BorutaPy
from yikit.tools import is_notebook
if is_notebook():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

import optuna
from joblib import Parallel, delayed

import numpy as np
import pandas as pd

__all__ = [
    'NNRegressor',
    'GBDTRegressor',
    'LinearModelRegressor',
    'SupportVectorRegressor',
    'EnsembleRegressor',
    'Objective'
]


class NNRegressor(BaseEstimator, RegressorMixin):
    def __init__(self, input_dropout = 0.0, hidden_layers = 3, hidden_units = 96, hidden_activation = 'relu', hidden_dropout = 0.2, batch_norm = 'before_act', optimizer_type = 'adam', lr = 0.001, batch_size = 64, l = 0.01, random_state = None, epochs = 200, patience = 20, progress_bar = True, scale = True):
        try:
            import keras
        except ModuleNotFoundError as e:
            sys.stdout.write(e)
            raise ModuleNotFoundError('If you want to use this module, please install keras.')

        self.input_dropout = input_dropout
        self.hidden_layers = hidden_layers
        self.hidden_units = hidden_units
        self.hidden_activation = hidden_activation
        self.hidden_dropout = hidden_dropout
        self.batch_norm = batch_norm
        self.optimizer_type = optimizer_type
        self.lr = lr
        self.batch_size = batch_size
        self.l = l
        self.random_state = random_state
        self.epochs = epochs
        self.patience = patience
        self.progress_bar = progress_bar
        self.scale = scale

    def fit(self, X, y):
        from keras.callbacks import EarlyStopping
        from keras.layers.advanced_activations import ReLU, PReLU
        from keras.layers import Dense, Dropout
        from keras.layers.normalization import BatchNormalization
        from keras.models import Sequential
        from keras.optimizers import SGD, Adam
        from keras.regularizers import l2
        from keras.backend import clear_session
        
        # 入力されたXとyが良い感じか判定（サイズが適切かetc)
        X, y = check_X_y(X, y)

        '''
        sklearn/utils/estimator_checks.py:3063:
        FutureWarning: As of scikit-learn 0.23, estimators should expose a n_features_in_ attribute, 
        unless the 'no_validation' tag is True.
        This attribute should be equal to the number of features passed to the fit method.
        An error will be raised from version 1.0 (renaming of 0.25) when calling check_estimator().
        See SLEP010: https://scikit-learn-enhancement-proposals.readthedocs.io/en/latest/slep010/proposal.html
        '''
        self.n_features_in_ = X.shape[1]    # check_X_yのあとでないとエラーになりうる．

        # check_random_state
        self.rng_ = check_random_state(self.random_state)

        # 標準化
        if self.scale:
            self.scaler_X_ = StandardScaler()
            X_ = self.scaler_X_.fit_transform(X)

            self.scaler_y_ = StandardScaler()
            y_ = self.scaler_y_.fit_transform(np.array(y).reshape(-1, 1)).flatten()
        else:
            X_ = X
            y_ = y
        

        # なぜか本当によくわかんないけど名前が一緒だと怒られるのでそれをずらすためのやつをつくる
        j = int(10E5)

        # プログレスバー
        if self.progress_bar:
            pbar = tqdm(total = 1 + self.hidden_layers + 6)
        self.estimator_ = Sequential()

        # 入力層
        self.estimator_.add(Dropout(self.input_dropout, input_shape = (self.n_features_in_,), seed = self.rng_.randint(2 ** 32), name = 'Dropout_' + str(j)))
        if self.progress_bar:
            pbar.update(1)


        # 中間層
        for i in range(self.hidden_layers):
            self.estimator_.add(Dense(units = self.hidden_units, kernel_regularizer = l2(l = self.l), name = 'Dense_' + str(j+i+1)))  # kernel_regularizer: 過学習対策
            if self.batch_norm == 'before_act':
                self.estimator_.add(BatchNormalization(name = 'BatchNormalization' + str(j+i+1)))
            
            # 活性化関数
            if self.hidden_activation == 'relu':
                self.estimator_.add(ReLU(name = 'Re_Lu_' + str(j+i+1)))
            elif self.hidden_activation == 'prelu':
                self.estimator_.add(PReLU(name = 'PRe_Lu_' + str(j+i+1)))
            else:
                raise NotImplementedError

            self.estimator_.add(Dropout(self.hidden_dropout, seed = self.rng_.randint(2 ** 32), name = 'Dropout_' + str(j+i+1)))

            # プログレスバー
            if self.progress_bar:
                pbar.update(1)
        
        # 出力層
        self.estimator_.add(Dense(1, activation = 'linear', name = 'Dense_' + str(j+self.hidden_layers+1)))

        # optimizer
        if self.optimizer_type == 'adam':
            optimizer_ = Adam(lr = self.lr, beta_1 = 0.9, beta_2 = 0.999, decay = 0.0)
        elif self.optimizer_type == 'sgd':
            optimizer_ = SGD(lr = self.lr, decay = 1E-6, momentum = 0.9, nesterov = True)
        else:
            raise NotImplementedError
        
        # 目的関数，評価指標などの設定
        self.estimator_.compile(loss = 'mean_squared_error', optimizer = optimizer_, metrics=['mse', 'mae'])

        # プログレスバー
        if self.progress_bar:
            pbar.update(1)

        # 変数の定義
        self.early_stopping_ = EarlyStopping(patience = self.patience, restore_best_weights = True)
        self.validation_split_ = 0.2

        # fit
        self.estimator_.fit(X_, y_, validation_split = self.validation_split_, epochs = self.epochs, batch_size = self.batch_size, callbacks = [self.early_stopping_], verbose = 0)

        # プログレスバー
        if self.progress_bar:
            pbar.update(5)
            pbar.close()

        return self

    def predict(self, X):
        # fitが行われたかどうかをインスタンス変数が定義されているかで判定（第二引数を文字列ではなくてリストで与えることでより厳密に判定可能）
        check_is_fitted(self, 'estimator_')

        # 入力されたXが妥当か判定
        X = check_array(X)

        if self.scale:
            X_ = self.scaler_X_.transform(X)
            y_pred_ = self.estimator_.predict(X_)
            y_pred_ = self.scaler_y_.inverse_transform(y_pred_).flatten()
        else:
            X_ = X
            y_pred_ = self.estimator_.predict(X_).flatten()
        return y_pred_
    


class GBDTRegressor(RegressorMixin, BaseEstimator):
    def __init__(self, boosting_type='gbdt', num_leaves=31, max_depth=-1, learning_rate=0.1, n_estimators=100, subsample_for_bin=200000, objective=None, class_weight=None, min_split_gain=0.0, min_child_weight=0.001, min_child_samples=20, subsample=1.0, subsample_freq=0, colsample_bytree=1.0, reg_alpha=0.0, reg_lambda=0.0, random_state=None, n_jobs=-1, silent=True, importance_type='split', **kwargs):
        # self.hoge = hogeとしなければいけない．つまりself.fuga = hogeだと怒られる
        self.boosting_type = boosting_type
        self.num_leaves = num_leaves
        self.max_depth = max_depth
        self.learning_rate = learning_rate
        self.n_estimators = n_estimators
        self.subsample_for_bin = subsample_for_bin
        self.objective = objective
        self.class_weight = class_weight
        self.min_split_gain = min_split_gain
        self.min_child_weight = min_child_weight
        self.min_child_samples = min_child_samples
        self.subsample = subsample
        self.subsample_freq = subsample_freq
        self.colsample_bytree = colsample_bytree
        self.reg_alpha = reg_alpha
        self.reg_lambda = reg_lambda
        self.random_state = random_state
        self.n_jobs = n_jobs
        self.silent = silent
        self.importance_type = importance_type
        if len(kwargs):
            self.kwargs = kwargs

    def fit(self, X, y):
        try:
            kwargs = self.kwargs
        except:
            kwargs = {}

        # check_random_state
        self.rng_ = check_random_state(self.random_state)

        # fitしたあとに確定する値は変数名 + '_' としなければならない．
        self.estimator_ = LGBMRegressor(boosting_type = self.boosting_type,
            num_leaves = self.num_leaves,
            max_depth = self.max_depth,
            learning_rate = self.learning_rate,
            n_estimators = self.n_estimators,
            subsample_for_bin = self.subsample_for_bin,
            objective = self.objective,
            class_weight = self.class_weight,
            min_split_gain = self.min_split_gain,
            min_child_weight = self.min_child_weight,
            min_child_samples = self.min_child_samples,
            subsample = self.subsample,
            subsample_freq = self.subsample_freq,
            colsample_bytree = self.colsample_bytree,
            reg_alpha = self.reg_alpha,
            reg_lambda = self.reg_lambda,
            random_state = self.rng_,
            n_jobs = self.n_jobs,
            silent = self.silent,
            importance_type = self.importance_type,
            **kwargs
        )

        # 入力されたXとyが良い感じか判定（サイズが適切かetc)
        X, y = check_X_y(X, y)

        '''
        sklearn/utils/estimator_checks.py:3063:
        FutureWarning: As of scikit-learn 0.23, estimators should expose a n_features_in_ attribute, 
        unless the 'no_validation' tag is True.
        This attribute should be equal to the number of features passed to the fit method.
        An error will be raised from version 1.0 (renaming of 0.25) when calling check_estimator().
        See SLEP010: https://scikit-learn-enhancement-proposals.readthedocs.io/en/latest/slep010/proposal.html
        '''
        self.n_features_in_ = X.shape[1]    # check_X_yのあとでないとエラーになりうる．

        X_train, X_test, y_train, y_test = train_test_split(X, y, random_state = self.rng_, test_size = 0.2)

        self.estimator_.fit(X, y, eval_set = [(X_test, y_test)], eval_metric = ['mse', 'mae'], early_stopping_rounds = 20, verbose = False)
        self.feature_importances_ = self.estimator_.feature_importances_

        # 慣例と聞いたはずなのにこれをreturnしないと怒られる．審査が厳しい．
        return self

    def predict(self, X):
        # fitが行われたかどうかをインスタンス変数が定義されているかで判定（第二引数を文字列ではなくてリストで与えることでより厳密に判定可能）
        check_is_fitted(self, 'estimator_')

        # 入力されたXが妥当か判定
        X = check_array(X)

        # 予測結果を返す
        return self.estimator_.predict(X)


# スケールの概念が入っていないので，それらを内包したscikit-learn準拠モデルを自分で定義する必要がある．
class EnsembleRegressor(BaseEstimator, RegressorMixin):
    def __init__(self, estimators = (RandomForestRegressor(),), method = 'blending', cv = 5, n_jobs = -1, random_state = None, scoring = 'neg_mean_squared_error', verbose = 0, boruta = True, opt = True):
        '''
        Parameters
        ----------
        estimators: 1-d list, default = (RandomForestRegressor(), )
            List of estimators to ensemble.

        method: {'blending', 'average', 'stacking'}, default = 'blending'
            How to ensemble.

        cv: int or callable, default = 5

        n_jobs: int, default = -1

        random_state: None, int or callable, default = None

        scoring: str, callable or list, default = 'neg_mean_squared_error'
            https://scikit-learn.org/stable/modules/model_evaluation.html

        verbose: int, default = 0

        boruta: bool, default = True
            Do boruta or not.

        opt: bool, default = True
            Do hyperparameter optimization or not.
        '''
        self.estimators = estimators
        self.method = method
        self.cv = cv
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.scoring = scoring
        self.verbose = verbose
        self.boruta = boruta
        self.opt = opt

    def fit(self, X, y):
        # よく使うので変数化
        self.n_estimators_ = len(self.estimators)

        # check_X_y
        X, y = check_X_y(X, y)

        '''
        sklearn/utils/estimator_checks.py:3063:
        FutureWarning: As of scikit-learn 0.23, estimators should expose a n_features_in_ attribute, 
        unless the 'no_validation' tag is True.
        This attribute should be equal to the number of features passed to the fit method.
        An error will be raised from version 1.0 (renaming of 0.25) when calling check_estimator().
        See SLEP010: https://scikit-learn-enhancement-proposals.readthedocs.io/en/latest/slep010/proposal.html
        '''
        self.n_features_in_ = X.shape[1]    # check_X_yのあとでないとエラーになりうる．
        
        # check_random_state
        rng_ = check_random_state(self.random_state)
        
        # isRegressor
        if sum([is_regressor(estimator) for estimator in self.estimators]) != self.n_estimators_:
            raise ValueError

        # check_cv
        cv_ = check_cv(self.cv, y = y, classifier = False)

        # check_scoring
        estimator = self.estimators[0]
        if callable(self.scoring):
            scorers = self.scoring
        elif self.scoring is None or isinstance(self.scoring, str):
            scorers = check_scoring(estimator = estimator)
        else:
            # 0.24.1のコードだと辞書を返すことになっているが，0.23.2ではtupleが返ってきてしまう？
            scorers = _check_multimetric_scoring(estimator, self.scoring)
            if isinstance(scorers, tuple):
                scorers = scorers[0]

        # 並列処理する部分を関数化
        def _f(i_train, i_test):
            X_train, X_test, y_train, y_test = X[i_train], X[i_test], y[i_train], y[i_test]

            if self.boruta:
                # 特徴量削減
                feature_selector_ = BorutaPy(estimator = RandomForestRegressor(n_jobs = -1, random_state = rng_), random_state = rng_, max_iter = 300, verbose = self.verbose)
                feature_selector_.fit(X_train, y_train)
                
                # 抽出
                support_ = feature_selector_.get_support()
                X_train_selected = feature_selector_.transform(X_train)
                X_test_selected = feature_selector_.transform(X_test)
            else:   # borutaしない場合でもresultsに組み込まれるので変数を定義しておく．
                support_ = np.ones(X_train.shape[1], dtype = np.bool)
                X_train_selected = X_train
                X_test_selected = X_test

            # verbose
            if self.verbose == 0:
                optuna.logging.disable_default_handler()

            results_estimators = []
            for estimator in self.estimators:
                if self.opt:
                    # ハイパーパラメータ（scoringで最初にしていしたやつで最適化）
                    objective = Objective(estimator, X_train_selected, y_train, cv = cv_, random_state = rng_, scoring = scorers.values()[0] if isinstance(scorers, dict) else scorers)
                    sampler = optuna.samplers.TPESampler(seed = rng_.randint(2 ** 32))

                    study = optuna.create_study(sampler = sampler, direction = 'maximize')
                    study.optimize(objective, n_trials = 100, n_jobs = 1)

                    # 最適化後のモデル
                    _best_estimator_ = objective.model(**objective.fixed_params_, **study.best_params)
                else:   # optunaしない場合でもresultsに組み込まれるので変数を定義しておく．
                    study = None
                    _best_estimator_ = clone(estimator)

                # fit
                _best_estimator_.fit(X_train_selected, y_train)

                # predict
                _y_pred_on_train = _best_estimator_.predict(X_train_selected)
                _y_pred_on_test = _best_estimator_.predict(X_test_selected)

                # score
                _train_scores = _score(_best_estimator_, X_train_selected, y_train, scorers)
                _test_scores = _score(_best_estimator_, X_test_selected, y_test, scorers)

                # importances
                _gi = _best_estimator_.feature_importances_ if 'feature_importances_' in dir(_best_estimator_) else None
                _pi = permutation_importance(_best_estimator_, X_test_selected, y_test, scoring = 'neg_mean_squared_error', n_repeats = 10, n_jobs = -1, random_state = rng_).importances

                # 予測結果をDataFrameにまとめる．
                _y_train = pd.DataFrame(np.hstack([y_train.reshape(-1, 1), _y_pred_on_train.reshape(-1, 1)]), columns = ['true', 'pred'], index = i_train)
                _y_test = pd.DataFrame(np.hstack([y_test.reshape(-1, 1), _y_pred_on_test.reshape(-1, 1)]), columns = ['true', 'pred'], index = i_test)

                results_estimators.append({
                        'estimators': _best_estimator_,
                        'params': _best_estimator_.get_params(),
                        'y_train': _y_train,
                        'y_test': _y_test,
                        'train_scores': _train_scores,
                        'test_scores': _test_scores,
                        'gini_importances': _gi,
                        'permutation_importances': _pi,
                        'studies': study,
                    })
            # verbose
            if self.verbose == 0:
                optuna.logging.disable_default_handler()

            # 出力結果をいい感じにする．←ここから
            ret = {}
            temp = {}
            for result_estimator in results_estimators: # それぞれのestimatorについて
                for k, v in result_estimator.items():   # その中の各々の値について
                    # スコア系かつそれらが複数指定されている場合だけ特別処理
                    if '_score' in k and isinstance(v, dict):
                        if k not in temp:
                            temp[k] = {}
                        for score_name, score in v.items():
                            if score_name in temp[k]:
                                temp[k][score_name].append(score)
                            else:
                                temp[k][score_name] = [score]
                    else:   # スコア系以外
                        if k in ret:
                            ret[k].append(v)
                        else:
                            ret[k] = [v]
            # scoreをためてるやつをBunch型に変換
            for k in temp:
                temp[k] = Bunch(**temp[k])
            
            # 返すように最終整形
            ret['support_'] = support_
            ret.update(temp)
            return ret
        
        # 上記で定義した_f関数どうしは互いに独立なので並列で処理する．
        parallel = Parallel(n_jobs = self.n_jobs, verbose = self.verbose)
        results = parallel(delayed(_f)(i_train, i_test) for i_train, i_test in cv_.split(X, y))
        # results = [_f(i_train, i_test) for i_train, i_test in cv_.split(X, y)]    # デバッグ用．並列しないでやる方法

        # データを整形
        self.results_ = {}
        for result in results:
            for k, v in result.items():
                if k in self.results_:
                    self.results_[k].append(v)
                else:
                    self.results_[k] = [v]

        # 扱いやすいようにBunch型に変換
        self.results_ = Bunch(**self.results_)

        # OOFの予測結果を取得
        dfs_y_oof_ = [pd.concat([lst[n] for lst in self.results_['y_test']], axis = 0).sort_index() for n in range(self.n_estimators_)]
        y_oof_ = pd.concat([df.loc[:, 'pred'] for df in dfs_y_oof_], axis = 1)
        y_oof_.columns = ['estimator{}'.format(n) for n in range(self.n_estimators_)]

        # *** ensemble ***
        # モデルがひとつのとき．
        if self.method == 'average' or self.n_estimators_ == 1:
            self.weights_ = None
        elif self.method == 'blending':
            # rmseで最適化（今後指定できるようにしてもいいかも．）
            def objective(trial):
                params = {'weight{0}'.format(i): trial.suggest_uniform('weight{0}'.format(i), 0, 1) for i in range(self.n_estimators_)}
                weights = np.array(list(params.values()))
                y_oof_ave = np.average(y_oof_, weights = weights, axis = 1)
                return mean_squared_error(y_oof_ave, y, squared = False)

            # optunaのログを非表示
            if self.verbose == 0:
                optuna.logging.disable_default_handler()

            # 重みの最適化
            sampler_ = optuna.samplers.TPESampler(seed = rng_.randint(2 ** 32))
            study = optuna.create_study(sampler = sampler_, direction = 'minimize') # 普通のRMSEなので．
            study.optimize(objective, n_trials = 100, n_jobs = 1)   # -1にするとなぜかバグるので．（そもそもそんなに重くないので1で．）

            # optunaのログを再表示
            if self.verbose == 0:
                optuna.logging.enable_default_handler()

            self.weights_ = np.array(list(study.best_params.values()), dtype = np.float64)
            self.weights_ /= np.sum(self.weights_)
        elif self.method == 'stacking':
            # 線形モデルの定義
            self.stacking_model_ = LinearRegression(n_jobs = self.n_jobs)
            self.stacking_model_.fit(y_oof_.values, y)
            # resultsに保存するために定義だけする．
            self.weights_ = None
        else:
            raise NotImplementedError

        # 重みを結果に保存
        self.results_['weights'] = self.weights_

        return self

    def predict(self, X):
        # fitが行われたかどうかをインスタンス変数が定義されているかで判定（第二引数を文字列ではなくてリストで与えることでより厳密に判定可能）
        check_is_fitted(self, 'results_')

        # 入力されたXが妥当か判定
        X = check_array(X)

        # 各予測モデルの予測結果をまとめる．(内包リストで得られる配列は3-Dベクトル (n_estimators_, cv, n_samples))
        y_preds_ = np.average(np.array([[estimators[n].predict(X[:, self.results_.support_[m]]) for m, estimators in enumerate(self.results_.estimators)] for n in range(self.n_estimators_)]), axis = 1).transpose()   # 同じ種類のやつは単純に平均を取る．
        
        if self.method in ('blending', 'average') or self.n_estimators_ == 1:
            y_pred_ = np.average(y_preds_, weights = self.weights_, axis = 1)
        elif self.method == 'stacking':
            y_pred_ = self.stacking_model_.predict(y_preds_)
        return y_pred_


class SupportVectorRegressor(BaseEstimator, RegressorMixin):
    def __init__(self, kernel = 'rbf', gamma = 'auto', tol = 0.01, C = 1.0, epsilon = 0.1, scale = True):
        self.kernel = kernel
        self.gamma = gamma
        self.tol = tol
        self.C = C
        self.epsilon = epsilon
        self.scale = scale

    def fit(self, X, y):
        # 入力されたXとyが良い感じか判定（サイズが適切かetc)
        X, y = check_X_y(X, y)

        '''
        sklearn/utils/estimator_checks.py:3063:
        FutureWarning: As of scikit-learn 0.23, estimators should expose a n_features_in_ attribute, 
        unless the 'no_validation' tag is True.
        This attribute should be equal to the number of features passed to the fit method.
        An error will be raised from version 1.0 (renaming of 0.25) when calling check_estimator().
        See SLEP010: https://scikit-learn-enhancement-proposals.readthedocs.io/en/latest/slep010/proposal.html
        '''
        self.n_features_in_ = X.shape[1]    # check_X_yのあとでないとエラーになりうる．

        if self.scale:
            self.scaler_X_ = StandardScaler()
            X_ = self.scaler_X_.fit_transform(X)

            self.scaler_y_ = StandardScaler()
            y_ = self.scaler_y_.fit_transform(np.array(y).reshape(-1, 1)).flatten()
        else:
            X_ = X
            y_ = y

        self.estimator_ = SVR(kernel=self.kernel, gamma=self.gamma, tol=self.tol, C=self.C, epsilon=self.epsilon)
        self.estimator_.fit(X_, y_)

        return self

    def predict(self, X):
        # fitが行われたかどうかをインスタンス変数が定義されているかで判定（第二引数を文字列ではなくてリストで与えることでより厳密に判定可能）
        check_is_fitted(self, 'estimator_')

        # 入力されたXが妥当か判定
        X = check_array(X)

        if self.scale:
            X_ = self.scaler_X_.transform(X)
        else:
            X_ = X

        y_pred_ = self.estimator_.predict(X_)
        if self.scale:
            y_pred_ = self.scaler_y_.inverse_transform(np.array(y_pred_).reshape(-1, 1)).flatten()
        
        return y_pred_

class LinearModelRegressor(BaseEstimator, RegressorMixin):
    def __init__(self, linear_model = 'ridge', alpha = 1.0, fit_intercept = True, max_iter = 1000, tol = 0.001, random_state = None):
        self.linear_model = linear_model
        self.alpha = alpha
        self.fit_intercept = fit_intercept
        self.max_iter = max_iter
        self.tol = tol
        self.random_state = random_state

    def fit(self, X, y):
        X, y = check_X_y(X, y)

        '''
        sklearn/utils/estimator_checks.py:3063:
        FutureWarning: As of scikit-learn 0.23, estimators should expose a n_features_in_ attribute, 
        unless the 'no_validation' tag is True.
        This attribute should be equal to the number of features passed to the fit method.
        An error will be raised from version 1.0 (renaming of 0.25) when calling check_estimator().
        See SLEP010: https://scikit-learn-enhancement-proposals.readthedocs.io/en/latest/slep010/proposal.html
        '''
        self.n_features_in_ = X.shape[1]    # check_X_yのあとでないとエラーになりうる．

        self.rng_ = check_random_state(self.random_state)

        # max_iterを引数に入れてるとこの変数ないとダメ！って怒られるから．
        self.n_iter_ = 1

        if self.linear_model == 'ridge':
            model_ = Ridge
        elif self.linear_model == 'lasso':
            model_ = Lasso
        else:
            raise NotImplementedError

        self.estimator_ = model_(alpha = self.alpha, fit_intercept = self.fit_intercept, max_iter = self.max_iter, tol = self.tol, random_state = self.rng_)
        self.estimator_.fit(X, y)
        return self
    
    def predict(self, X):
        check_is_fitted(self, 'estimator_')

        X = check_array(X)

        return self.estimator_.predict(X)

class Objective:
    def __init__(self, estimator, X, y, custom_params = lambda trial: {}, fixed_params = {}, cv = 5, random_state = None, scoring = None, n_jobs = None):
        """objective function of optuna.

        Parameters
        ----------
        estimator : sklearn-based estimator instance
            e.g. sklearn.ensemble.RandomForestRegressor()
        X : 2-d array
            features
        y : 1-d array
            target
        custom_params : func, optional
            If you want to do your own custom range of optimization, you can define it here with a function that returns a dictionary., by default lambda trial:{}
        fixed_params : dict, optional
            If you have a fixed variable, you can specify it in the dictionary., by default {}
        cv : int, KFold object, optional
            How to cross-validate, by default 5
        random_state : int or RandomState object, optional
            seed, by default None
        scoring : scorer object, str, etc., optional
            If you don't specify it, it will use the default evaluation function of sklearn., by default None
        n_jobs : int, optional
            parallel processing, by default None
        """
        self.estimator = estimator
        self.X, self.y = check_X_y(X, y)
        self.custom_params = custom_params
        self._fixed_params = fixed_params
        self.cv = check_cv(cv)
        self.rng = check_random_state(random_state)
        self.scoring = check_scoring(estimator, scoring)
        self.n_jobs = n_jobs

        # sampler
        self.sampler = optuna.samplers.TPESampler(seed = self.rng.randint(2 ** 32 - 1))

    def __call__(self, trial):
        if isinstance(self.estimator, NNRegressor):
            params_ = {
                'input_dropout': trial.suggest_uniform('input_dropout', 0.0, 0.3),
                'hidden_layers': trial.suggest_int('hidden_layers', 2, 4),
                'hidden_units' : trial.suggest_int('hidden_units', 32, 1024, 32),
                'hidden_activation' : trial.suggest_categorical('hidden_activation', ['prelu', 'relu']),
                'hidden_dropout' : trial.suggest_uniform('hidden_dropout', 0.2, 0.5),
                'batch_norm' : trial.suggest_categorical('batch_norm', ['before_act', 'no']),
                'optimizer_type' : trial.suggest_categorical('optimizer_type', ['adam', 'sgd']),
                'lr' : trial.suggest_loguniform('lr', 0.00001, 0.01),
                'batch_size' : trial.suggest_int('hidden_units', 32, 1024, 32),
                'l' : trial.suggest_loguniform('l', 1E-7, 0.1),
            }
            self.fixed_params_ = {
                'progress_bar': False,
                'random_state': self.rng,
            }
        elif isinstance(self.estimator, (GBDTRegressor, LGBMRegressor)):
            params_ = {
                'n_estimators' : trial.suggest_int('n_estimators', 10, 1000, log=True),
                # 'max_depth' : trial.suggest_int('n_estimators', 3, 9),    # num_leaves変えた方が良さそう．制約条件的に．
                'min_child_weight' : trial.suggest_loguniform('min_child_weight', 0.001, 10),
                'colsample_bytree' : trial.suggest_uniform('colsample_bytree', 0.6, 0.95),
                'subsample': trial.suggest_uniform('subsample', 0.6, 0.95),
                'num_leaves' : trial.suggest_int('num_leaves', 2 ** 3, 2 ** 9, log = True)
            }
            self.fixed_params_ = {
                'random_state' : self.rng,
                'n_jobs' : -1,
                'objective' : 'regression',
            }
        elif isinstance(self.estimator, RandomForestRegressor):
            # 最適化するべきパラメータ
            params_ = {
                'min_samples_split' : trial.suggest_int('min_samples_split', 2, 16),
                'max_depth' : trial.suggest_int('max_depth', 10, 100),
                'n_estimators' : trial.suggest_int('n_estimators', 10, 1000, log=True),
            }
            # 固定するパラメータ (外でも取り出せるようにインスタンス変数としてる．)
            self.fixed_params_ = {
                'random_state' : self.rng,
                'n_jobs' : -1,
            }
        elif isinstance(self.estimator, (SupportVectorRegressor, SVR)):
            # 最適化するべきパラメータ
            params_ = {
                'C' : trial.suggest_loguniform('C', 2 ** -5, 2 ** 10),
                'epsilon' : trial.suggest_loguniform('epsilon', 2 ** -10, 2 ** 0),
            }
            # 固定するパラメータ (外でも取り出せるようにインスタンス変数としてる．)
            self.fixed_params_ = {
                'gamma' : 'auto',
                'kernel' : 'rbf'
            }
        elif isinstance(self.estimator, LinearModelRegressor):
            # 最適化するべきパラメータ
            params_ = {
                'linear_model' : trial.suggest_categorical('linear_model', ['ridge', 'lasso']),
                'alpha' : trial.suggest_loguniform('alpha', 0.1, 10),
                'fit_intercept' : trial.suggest_categorical('fit_intercept', [True, False]),
                'max_iter' : trial.suggest_loguniform('max_iter', 100, 10000),
                'tol' : trial.suggest_loguniform('tol', 0.0001, 0.01),
            }
            # 固定するパラメータ (外でも取り出せるようにインスタンス変数としてる．)
            self.fixed_params_ = {
                'random_state' : self.rng,
            }
        elif isinstance(self.estimator, MLPRegressor):
            # 最適化するべきパラメータ
            params_ = {
                'hidden_layer_sizes': trial.suggest_int('hidden_layer_sizes', 50, 300),
                'alpha': trial.suggest_loguniform('alpha', 1e-5, 1e-3),
                'learning_rate_init': trial.suggest_loguniform('learning_rate_init', 1e-5, 1e-3),
            }
            # 固定するパラメータ (外でも取り出せるようにインスタンス変数としてる．)
            self.fixed_params_ = {
                'random_state' : self.rng,
            }
        elif isinstance(self.estimator, NGBRegressor):
            # 最適化するべきパラメータ
            params_ = {
                'Base': DecisionTreeRegressor(
                    max_depth =  trial.suggest_int('_max_depth', 2, 100),
                    criterion = trial.suggest_categorical('_criterion', ['mse', 'friedman_mse']),
                    random_state = self.rng,
                ),
                'n_estimators' : trial.suggest_int('n_estimators', 10, 1000, log=True),
                'minibatch_frac': trial.suggest_uniform('minibatch_frac', 0.5, 1.0),
            }
            # 固定するパラメータ (外でも取り出せるようにインスタンス変数としてる．)
            self.fixed_params_ = {
                'random_state' : self.rng,
            }
        elif self.custom_params(trial):
            params_ = self.custom_params(trial)
            self.fixed_params_ = {} # あとで加えるので空でOK．
        else:
            raise NotImplementedError('{0}'.format(self.estimator))

        # もしfixed_paramsを追加で指定されたらそれを取り入れる
        self.fixed_params_.update(self._fixed_params)

        self.model = type(self.estimator)
        # self.estimator_ = self.model(**params_, **self.fixed_params_)
        self.estimator_ = clone(self.estimator)
        self.estimator_.set_params(
            **params_,
            **self.fixed_params_,
        )

        parallel = Parallel(n_jobs = self.n_jobs)
        results = parallel(
            delayed(_fit_and_score)(
                clone(self.estimator_), self.X, self.y, self.scoring, train, test, 0, dict(**self.fixed_params_, **params_), None
            )
        for train, test in self.cv.split(self.X, self.y))
        return np.mean([d['test_scores'] for d in results]) # scikit-learn>=0.24.1

    def get_best_estimator(self, study):
        best_params_ = self.get_best_params(study)
        return self.model(**best_params_)

    def get_best_params(self, study):
        if isinstance(self.estimator_, NGBRegressor):
            dt_best_params_ = {}
            best_params_ = {}
            for k, v in study.best_params.items():
                if k[0] == '_':
                    dt_best_params_[k[1:]] = v
                else:
                    best_params_[k] = v
            else:
                if 'random_state' in self.fixed_params_:
                    dt_best_params_['random_state'] = self.fixed_params_['random_state']
                best_params_['Base'] = DecisionTreeRegressor(**dt_best_params_)
        else:
            best_params_ = study.best_params
        best_params_.update(**self.fixed_params_)
        return best_params_



if __name__ == '__main__':
    pass



'''
Copyright (c) 2021 yu9824

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import sys

def is_notebook():
    """the environment is notebook or not.
    reference: https://blog.amedama.jp/entry/detect-jupyter-env

    Returns
    -------
    bool
        is_notebook or not.
    """
    if 'get_ipython' not in globals():
        return 'ipykernel' in sys.modules
        # Python shell
        # return False
    env_name = get_ipython().__class__.__name__
    if env_name == 'TerminalInteractiveShell':
        # IPython shell
        return False
    return True

'''
例
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm
'''

if __name__ == '__main__':
    pass


from ngboost import NGBRegressor, NGBClassifier

def log_likelihood(estimator, X, y):
    if isinstance(estimator, (NGBRegressor, NGBClassifier)):
        return - estimator.score(X, y)  # LL (- NLL)
    else:
        raise TypeError("'estimator' is not {0} or {1}".format(NGBRegressor, NGBClassifier))



'''
Copyright (c) 2021 yu9824

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import optuna
import seaborn as sns
import sys
from decimal import Decimal
from math import ceil
import warnings
from ngboost.distns import Distn
from ngboost import NGBClassifier, NGBRegressor
from lightgbm import LGBMClassifier, LGBMRegressor
import platform
from sklearn.utils import Bunch

from sklearn.utils.validation import check_array
from yikit.tools import is_notebook

if is_notebook():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

COLORS = Bunch(
    train = '#283655',
    test = '#cf3721',
    val = '#cf3721',
)

def get_learning_curve_optuna(study, loc = 'best', fontfamily='Helvetica', return_axis = False):
    """get_leraning_curve

    Parameters
    ----------
    study : optuna.study.Study
        
    loc : str, optional
        legend's location, by default 'best'
    fontfamily : str, optional
        fontfamily, by default 'Helvetica'
    return_axis : bool, optional
        return axis or not, by default False

    Returns
    -------
    if return_axis is True:
        tuple (matplotlib.pyplot.figure, matplotlib.pyplot.axis)
    else:
        matplotlib.pyplot.figure
    """
    # 2: maximize, 1: minimize
    if study.direction == optuna.study.StudyDirection(1):
        min_or_max = min
        temp = float('inf')
    elif study.direction == optuna.study.StudyDirection(2):
        min_or_max = max
        temp = - float('inf')
    df_values = pd.DataFrame([trial.value for trial in study.trials], columns = ['value'])
    
    best_values = []
    for v in df_values.loc[:, 'value']:
        temp = min_or_max(temp, v)
        best_values.append(temp)
    df_values['best_value'] = best_values

    _font_settings(fontfamily=fontfamily)

    fig = plt.figure(facecolor = 'white')
    ax = fig.add_subplot(111)

    ax.scatter(df_values.index, df_values['value'], s = 10, c = '#00293c', label = 'Objective Value')
    ax.plot(df_values.index, df_values['best_value'], c = '#f62a00', zorder = 0, label = 'Best Value')

    ax.set_xlabel('Trials')
    ax.set_ylabel('Objective Values')

    ax.legend(facecolor = '#f0f0f0', edgecolor = 'None', fontsize = 10, loc = loc)

    fig.tight_layout()
    if return_axis:
        return fig, ax
    else:
        return fig

class SummarizePI:
    def __init__(self, importances):
        """Summarize permutation importances

        Parameters
        ----------
        importances : pandas.DataFrame
            index: features
            columns: n_repeats
        """
        self.importances = importances

    def get_figure(self, fontfamily='Helvetica'):
        # 平均をとる．
        imp = self.importances.mean(axis = 1)

        # 「規格化」したのち，大きい順に並べ替えて，sns.bairplotのためにtranspose()
        df_imp = pd.DataFrame(imp / np.sum(imp), columns = ['importances']).sort_values('importances', ascending=False).transpose()

        # レイアウトについて
        _font_settings(fontfamily=fontfamily)

        # 重要度の棒グラフを描画
        self.fig = plt.figure(facecolor = 'white')
        self.ax = self.fig.add_subplot(111)

        sns.barplot(data = df_imp, ax = self.ax, orient = 'h')

        self.fig.tight_layout()

        return self.fig, self.ax

    def get_data(self):
        pass

def get_dist_figure(y_dist, y_true = None, keep_y_range = True, return_axis = False, verbose = True, titles = [], fontfamily='Helvetica'):
    """get distribution figure.

    Parameters
    ----------
    y_dist : ngboost.distns.Distn object
        generated by ngb.pred_dist(X)
    y_true : 1d array, optional
        [description], by default None
    keep_y_range : bool, optional
        keep y range equal or not, by default True
    return_axis : bool, optional
        return axis or not, by default False
    verbose : bool, optional
        drwaing progress bar or not, by default True
    titles : 1d list, optional
        titles whose lengh == n_samples, by default []
    fontfamily : str, optional
        fontfamily, by default 'Helvetica'

    Returns
    -------
    matplotlib.pyplot.Figure | (matplotlib.pyplot.Figure, matplotlib.pyplot.Axis)
    """
    # check
    if not isinstance(y_dist, Distn):
        raise TypeError('`y_dist` is not {0} object'.format(Distn.__name__))
    y_pred = y_dist.mean()
    y_pred = check_array(y_pred, ensure_2d=False)
    n_samples = len(y_pred)
    if len(titles) not in (0, n_samples):
        raise ValueError("`titles`'s lengh must be 0 or n_samples({}).".format(n_samples))

    offset = np.ptp(y_pred) * 0.05
    y_range = np.linspace(min(y_pred)-offset, max(y_pred)+offset, 200).reshape((-1, 1))
    dist_values = y_dist.pdf(y_range).transpose()

    n_rows = ceil(Decimal(n_samples).sqrt())
    n_cols = n_samples // n_rows + int(n_samples % n_rows > 0)

    _font_settings(fontfamily=fontfamily)
    fig, axes = plt.subplots(n_rows, n_cols, facecolor='white', dpi=72, figsize=(6.4*n_cols, 4.8*n_rows))
    
    if verbose:
        pbar = tqdm(total=n_cols * n_rows + 1, desc='Drawing distribution figure')
    # 一つずつ分布を書いていく．
    for idx in range(n_samples):
        ax = axes[idx//n_cols][idx%n_cols]
        ax.plot(y_range, dist_values[idx], c = '#022C5E')
        
        prob_max_temp = max(dist_values[idx])   # このidxにおける確率密度の最大値
        if y_true is not None:
            y_true = check_array(y_true, ensure_2d=False)
            ax.vlines(y_true[idx], 0, prob_max_temp, "#ab4e15", label="ground truth")
        ax.vlines(y_pred[idx], 0, prob_max_temp, "#dc143c", label="pred")
        ax.set_ylabel('Probability density')
        ax.legend(loc="best", facecolor='#f0f0f0', edgecolor='None')
        if titles:
            ax.set_title("{0}".format(titles[idx]))
        else:
            ax.set_title("idx: {0}".format(idx))
            
        
        ax.set_xlim(y_range[0], y_range[-1])
        if keep_y_range:
            ax.set_ylim(None, np.max(dist_values) * 1.05)

        if verbose:
            pbar.update(1)
    # 余ったところを消す．
    for idx in range(n_samples, n_rows * n_cols):
        ax = axes[idx//n_cols][idx%n_cols]
        ax.axis("off")
        if verbose:
            pbar.update(1)

    fig.tight_layout()

    if verbose:
        pbar.update(1)
        pbar.set_description(desc = 'Completed')
        pbar.close()
    if return_axis:
        return fig, axes
    else:
        return fig

def is_correct_dist(y_pred, y_dist):
    b = np.allclose(y_dist.mean(), y_pred)
    if not b:
        warnings.warn('`y_dist.mean()` and `y_pred` is not close.')
    return b

def select_fontfamily():
    if platform.system() == 'Darwin':
        return 'Helvetica'
    else:   # Window or Linux
        return 'DejaVu Sans'

def _font_settings(fontfamily='Helvetica', fontsize=13):
    plt.rcParams['font.size'] = fontsize
    plt.rcParams['font.family'] = fontfamily


def get_learning_curve_gb(estimator, fontfamily = 'Helvetica', return_axis = False):
    # check_estimator
    if isinstance(estimator, (NGBClassifier, NGBRegressor)):
        evals_result = estimator.evals_result
    elif isinstance(estimator, (LGBMClassifier, LGBMRegressor)):
        evals_result = estimator.evals_result_
    else:
        raise TypeError(estimator.__class__.__name__)

    # font
    _font_settings(fontfamily=fontfamily)

    # generate figure
    fig, ax = plt.subplots(1, 1, facecolor='white', dpi=72)

    # plot
    for data_name, result in evals_result.items():
        if data_name not in COLORS:
            raise ValueError('{} is not in `COLOR`.'.format(data_name))
        score_name = list(result.keys())[0]
        score = result[score_name]
        ax.plot(range(len(score)), score, label = data_name, c = COLORS[data_name])
    
    # label
    ax.set_xlabel('n_iter')
    ax.set_ylabel(score_name)

    # legend
    ax.legend(facecolor='#f0f0f0', edgecolor='None', fontsize=10)
    # tight_layout
    fig.tight_layout()
    if return_axis:
        return fig, ax
    else:
        return fig




'''
Copyright © 2021 yu9824

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from boruta import BorutaPy
from sklearn.utils import check_X_y
from sklearn.utils import check_random_state
from sklearn.utils import shuffle
from sklearn.utils.validation import check_is_fitted
import numpy as np
import pandas as pd
from scipy.stats import pearsonr
import sys
import warnings

from yikit.tools import is_notebook


class BorutaPy(BorutaPy):
    def __init__(self, estimator, n_estimators='auto', perc='auto', alpha=0.05,
                 two_step=True, max_iter=100, random_state=None, verbose=1,
                 max_shuf=10000):
        """
        This docstring is modified from and uses parts of scikit-learn-contrib/boruta_py, which is a class inheritor under the BSD 3 clause license.
        https://github.com/scikit-learn-contrib/boruta_py/blob/master/boruta/boruta_py.py

        Improved Python implementation of the Boruta R package.
        The improvements of this implementation include:
        - Faster run times:
            Thanks to scikit-learn's fast implementation of the ensemble methods.
        - Scikit-learn like interface:
            Use BorutaPy just like any other scikit learner: fit, fit_transform and
            transform are all implemented in a similar fashion.
        - Modularity:
            Any ensemble method could be used: random forest, extra trees
            classifier, even gradient boosted trees.
        - Two step correction:
            The original Boruta code corrects for multiple testing in an overly
            conservative way. In this implementation, the Benjamini Hochberg FDR is
            used to correct in each iteration across active features. This means
            only those features are included in the correction which are still in
            the selection process. Following this, each that passed goes through a
            regular Bonferroni correction to check for the repeated testing over
            the iterations.
        - Percentile:
            Instead of using the max values of the shadow features the user can
            specify which percentile to use. This gives a finer control over this
            crucial parameter. For more info, please read about the perc parameter.
        - Automatic tree number:
            Setting the n_estimator to 'auto' will calculate the number of trees
            in each itartion based on the number of features under investigation.
            This way more trees are used when the training data has many feautres
            and less when most of the features have been rejected.
        - Ranking of features:
            After fitting BorutaPy it provides the user with ranking of features.
            Confirmed ones are 1, Tentatives are 2, and the rejected are ranked
            starting from 3, based on their feautre importance history through
            the iterations.
        We highly recommend using pruned trees with a depth between 3-7.
        For more, see the docs of these functions, and the examples below.
        Original code and method by: Miron B Kursa, https://m2.icm.edu.pl/boruta/
        Boruta is an all relevant feature selection method, while most other are
        minimal optimal; this means it tries to find all features carrying
        information usable for prediction, rather than finding a possibly compact
        subset of features on which some classifier has a minimal error.
        Why bother with all relevant feature selection?
        When you try to understand the phenomenon that made your data, you should
        care about all factors that contribute to it, not just the bluntest signs
        of it in context of your methodology (yes, minimal optimal set of features
        by definition depends on your classifier choice).
        Parameters
        ----------
        estimator : object
            A supervised learning estimator, with a 'fit' method that returns the
            feature_importances_ attribute. Important features must correspond to
            high absolute values in the feature_importances_.
        n_estimators : int or string, default = 1000
            If int sets the number of estimators in the chosen ensemble method.
            If 'auto' this is determined automatically based on the size of the
            dataset. The other parameters of the used estimators need to be set
            with initialisation.
        perc : int | 'auto', default = 'auto'
            Instead of the max we use the percentile defined by the user, to pick
            our threshold for comparison between shadow and real features. The max
            tend to be too stringent. This provides a finer control over this. The
            lower perc is the more false positives will be picked as relevant but
            also the less relevant features will be left out. The usual trade-off.
            The default is essentially the vanilla Boruta corresponding to the max.
        alpha : float, default = 0.05
            Level at which the corrected p-values will get rejected in both
            correction steps.
        two_step : Boolean, default = True
            If you want to use the original implementation of Boruta with Bonferroni
            correction only set this to False.
        max_iter : int, default = 100
            The number of maximum iterations to perform.
        random_state : int, RandomState instance or None; default=None
            If int, random_state is the seed used by the random number generator;
            If RandomState instance, random_state is the random number generator;
            If None, the random number generator is the RandomState instance used
            by `np.random`.
        verbose : int, default=1
            Controls verbosity of output:
            - 0: no output
            - 1: displays iteration number | if tqdm package exists, progress bar
            - 2: which features have been selected already
        max_shuf : int, default=10000
            How many times to calculate the Pearson coefficient when `perc` is 'auto'.

        Attributes
        ----------
        n_features_ : int
            The number of selected features.
        support_ : array of shape [n_features]
            The mask of selected features - only confirmed ones are True.
        support_weak_ : array of shape [n_features]
            The mask of selected tentative features, which haven't gained enough
            support during the max_iter number of iterations..
        ranking_ : array of shape [n_features]
            The feature ranking, such that ``ranking_[i]`` corresponds to the
            ranking position of the i-th feature. Selected (i.e., estimated
            best) features are assigned rank 1 and tentative features are assigned
            rank 2.
        importance_history_ : array-like, shape [n_features, n_iters]
            The calculated importance values for each feature across all iterations.  
        Examples
        --------
        
        import pandas as pd
        from sklearn.ensemble import RandomForestClassifier
        from boruta import BorutaPy
        
        # load X and y
        # NOTE BorutaPy accepts numpy arrays only, hence the .values attribute
        X = pd.read_csv('examples/test_X.csv', index_col=0).values
        y = pd.read_csv('examples/test_y.csv', header=None, index_col=0).values
        y = y.ravel()
        
        # define random forest classifier, with utilising all cores and
        # sampling in proportion to y labels
        rf = RandomForestClassifier(n_jobs=-1, class_weight='balanced', max_depth=5)
        
        # define Boruta feature selection method
        feat_selector = BorutaPy(rf, n_estimators='auto', verbose=2, random_state=1)
        
        # find all relevant features - 5 features should be selected
        feat_selector.fit(X, y)
        
        # check selected features - first 5 features are selected
        feat_selector.support_
        
        # check ranking of features
        feat_selector.ranking_
        
        # call transform() on X to filter it down to selected features
        X_filtered = feat_selector.transform(X)
        References
        ----------
        [1] Kursa M., Rudnicki W., "Feature Selection with the Boruta Package"
            Journal of Statistical Software, Vol. 36, Issue 11, Sep 2010
        """
        self.max_shuf = max_shuf
        super().__init__(estimator=estimator, n_estimators=n_estimators, perc=perc, alpha=alpha,
                 two_step=two_step, max_iter=max_iter, random_state=random_state, verbose=verbose)
        self.random_state = check_random_state(self.random_state)
        if verbose > 0:
            try:
                import tqdm
            except ImportError as e:
                mess = '{}\nIf exists, a progress bar can be displayed.'.format(e)
                warnings.warn(mess)
                self._flag_tqdm = False
            else:
                self._flag_tqdm = True
        else:
            self._flag_tqdm = False
    
    def fit(self, X, y):
        """
        This docstring uses the same one as scikit-learn-contrib/boruta_py, which is a class inheritor under the BSD 3 clause license.
        https://github.com/scikit-learn-contrib/boruta_py/blob/master/boruta/boruta_py.py

        Fits the Boruta feature selection with the provided estimator.
        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            The training input samples.
        y : array-like, shape = [n_samples]
            The target values.
        """
        X, y = check_X_y(X, y)
        if self.perc == 'auto':
            self.perc = self._calc_auto_perc(X, y)
        if self._flag_tqdm and self.verbose == 1:
            if is_notebook():
                from tqdm.notebook import tqdm
            else:
                from tqdm import tqdm
            self.pbar = tqdm(total=self.max_iter, desc='BorutaPy')
        return self._fit(X, y)

    def get_support(self, weak=False):
        """get support

        Parameters
        ----------
        weak : bool, optional
            If set to true, the tentative features are also used to reduce X., by default False

        Returns
        -------
        support : array
        """
        check_is_fitted(self, 'support_')
        if weak:
            return self.support_weak_
        else:
            return self.support_
        
    
    def _calc_auto_perc(self, X, y):
        """
        This docstring is based on scikit-learn-contrib/boruta_py, which is a class inheritor under the BSD 3 clause license.
        https://github.com/scikit-learn-contrib/boruta_py/blob/master/boruta/boruta_py.py

        calculate `perc` if 'auto'

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            The training input samples.
        y : array-like, shape = [n_samples]
            The target values.

        Returns
        -------
        float
            calculated `perc`
        """
        if self.verbose > 0 and self._flag_tqdm:
            if is_notebook():
                from tqdm.notebook import trange
            else:
                from tqdm import trange
            _range = trange(self.max_shuf, desc = 'Calc r_ccmax')
        else:
            _range = range(self.max_shuf)
        
        # ランダムに並べ替えてどれくらい相関がでてしまうのかを調べ，自動で決める．
        self.pears_ = []    # 相関係数を足していくリスト
        for _ in _range:
            X_shuffled = shuffle(X, random_state = self.random_state)
            temp = [pearsonr(X_shuffled[:, i], y)[0] for i in range(X_shuffled.shape[1])]
            self.pears_.extend(temp)
        r_ccmax = max(self.pears_)
        return 100 * (1 - r_ccmax)
    
    def _print_results(self, dec_reg, _iter, flag):
        n_iter = str(_iter) + ' / ' + str(self.max_iter)
        n_confirmed = np.where(dec_reg == 1)[0].shape[0]
        n_rejected = np.where(dec_reg == -1)[0].shape[0]
        cols = ['Iteration: ', 'Confirmed: ', 'Tentative: ', 'Rejected: ']

        # still in feature selection
        if flag == 0:
            n_tentative = np.where(dec_reg == 0)[0].shape[0]
            content = map(str, [n_iter, n_confirmed, n_tentative, n_rejected])
            if self.verbose == 1:
                if self._flag_tqdm:
                    self.pbar.update()
                    output = None
                else:
                    output = cols[0] + n_iter
            elif self.verbose > 1:
                output = '\n'.join([x[0] + '\t' + x[1] for x in zip(cols, content)])

        # Boruta finished running and tentatives have been filtered
        else:
            n_tentative = np.sum(self.support_weak_)
            n_rejected = np.sum(~(self.support_|self.support_weak_))
            content = map(str, [n_iter, n_confirmed, n_tentative, n_rejected])
            result = '\n'.join([x[0] + '\t' + x[1] for x in zip(cols, content)])
            output = "\n\nBorutaPy finished running.\n\n" + result
            if self.verbose == 1 and self._flag_tqdm:
                self.pbar.update(self.max_iter - _iter + 1)
                self.pbar.close()
        
        if not(flag == 0 and self.verbose == 1 and self._flag_tqdm):
            sys.stdout.write(output + '\n')

if __name__ == '__main__':
    from pdb import set_trace
    



'''
Copyright © 2021 yu9824

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from sklearn.feature_selection import SelectorMixin # >= 0.23.2
from sklearn.base import BaseEstimator
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted

from scipy.stats import pearsonr
import pandas as pd
import numpy as np

from yikit.tools import is_notebook
if is_notebook():
    from tqdm.notebook import trange
else:
    from tqdm import trange

# from joblib import Parallel, delayed



class FilterSelector(SelectorMixin, BaseEstimator):
    def __init__(self, r = 0.9, alpha = 0.05):
        """Filter method (feature selection)

        Parameters
        ----------
        r : float, optional
            相関係数を高いとする閾値．これより高いと相関が高いとして，削除対象とする．, by default 0.9
        alpha : float, optional
            有意水準．これよりp値が低い時，相関係数の値が信頼に足ると判断．特別な理由がない限りはこれを変えないのがベター．, by default 0.05
        """
        self.r = r
        self.alpha = alpha

    def fit(self, X, y = None):
        """fit

        Parameters
        ----------
        X : pd.DataFrame or 2次元のnp.ndarray or 2次元のlist．
            すでに作成された特徴量空間．
        y : None, optional
            不要．, by default None
        """
        # checkかつnp.ndarrayに変換される．
        X = check_array(X)

        # よくつかうので列数を変数で保存
        n_features = X.shape[1]

        # 相関係数とそのp値を入れるリストの鋳型を作成
        # [i, j, 0]が相関係数， [i, j, 1]がp値
        self.corr_ = np.empty((n_features, n_features, 2), dtype = float)

        # i番目の要素は，i番目と相関係数がcorr以上かつp値が有意水準alpha未満のものの集合．これの鋳型を作成
        pair = [set() for _ in range(n_features)]

        for i in range(n_features):
            for j in range(i, n_features):
                if i == j:  # 一緒の時はそもそも相関係数1, p値0は自明なので例外処理
                    self.corr_[i, i] = [1.0, 0.0]
                else:
                    self.corr_[i, j] = list(pearsonr(X[:, i], X[:, j]))
                    self.corr_[j, i] = self.corr_[i, j]
                    if self.corr_[i, j, 0] > self.r and self.corr_[i, j, 1] < self.alpha:    # 相関係数が高く， かつp値が小さい (信頼に足る) とき．
                        # i, j番目の要素の集合にそれぞれj, iを追加
                        pair[i].add(j)
                        pair[j].add(i)

        # 何個相関係数が高い係数が存在するのかの値を求める．
        def _delete_recursive(pair, boolean = np.ones(n_features, dtype = bool)):
            # 相関係数が高いものがいくつあるのかのnp.ndarrayを作成
            order_pair = np.array([len(s) for s in pair])

            # これがすべて0のとき，もう削除は終わってるので終了処理
            if np.sum(order_pair) == 0:
                return boolean

            # 一番相関係数が高いペア数を多く持っているものの要素番号 (index) を取得
            i_max = np.argmax(order_pair)

            # 相関係数高いペアからその番号を削除
            pair = [p.difference({i_max}) for p in pair]

            # その集合は削除かつ，その特徴量はbooleanをFalseに．
            pair[i_max] = set()
            boolean[i_max] = False
            return _delete_recursive(pair, boolean)

        self.support_ = _delete_recursive(pair)

        return self

    def _get_support_mask(self):
        check_is_fitted(self, 'support_')
        return self.support_

if __name__ == '__main__':
    from pdb import set_trace
    from sklearn.datasets import load_boston

    data = load_boston()
    X = pd.DataFrame(data.data, columns = data.feature_names)
    # X = pd.concat([X, pd.Series(np.ones(X.shape[0]))], axis = 1)

    selector = FilterSelector()
    selector.fit(X)

    set_trace()




# 正規表現でtextファイルと同じやり方で読み込むのですべて手打ちする．

__version__      = '0.3.7'
__license__      = 'Apache License 2.0'
__author__       = 'yu9824'
__copyright__    = 'Copyright © 2021 yu9824'
__author_email__ = 'yu.9824.job@gmail.com'
__user_id__ = 'yu9824'
__app_name__ = 'yikit'
__url__ = 'http://github.com/yu9824/yikit/'

# __all__ = [
#     'models',
#     'tools',
#     'feature_selection',
#     ]


if __name__ == '__main__':
    from models import NNRegressor
    import pandas as pd
    
    from sklearn.utils.estimator_checks import check_estimator
    from sklearn.model_selection import train_test_split
    from sklearn.datasets import load_boston

    # 0.24からclassを入れる機能は削除された．
    # check_estimator(GBDTRegressor())
    check_estimator(NNRegressor())
    # check_estimator(EnsembleRegressor())
    # print(EnsembleRegressor)
    # check_estimator(SupportVectorRegressor())
    # print(SupportVectorRegressor)
    # check_estimator(LinearModelRegressor())
    # print(LinearModelRegressor)

    boston = load_boston()
    X = pd.DataFrame(boston['data'], columns = boston['feature_names'])
    y = pd.Series(boston['target'], name = 'PRICE')

    # objective = Objective(RandomForestRegressor(), X, y, scoring = 'neg_mean_squared_error')
    # trial = optuna.trial.FixedTrial({
    #     'min_samples_split': 2,
    #     'max_depth': 10,
    #     'n_estimators': 100,
    # })
    # print(objective(trial))

    X_train, X_test, y_train, y_test = train_test_split(X, y)

    # estimator = EnsembleRegressor(scoring = ['r2', 'neg_mean_squared_error'], random_state = 334, verbose = 1, boruta = True, opt = False, method = 'stacking')
    # estimator = EnsembleRegressor(estimators = [RandomForestRegressor(), LinearRegression()], scoring = ['r2', 'neg_mean_squared_error'], random_state = 334, boruta = False, opt = False, verbose = 1, method = 'stacking')
    # estimator = EnsembleRegressor(estimators = [RandomForestRegressor(), LGBMRegressor()], scoring = None, random_state = 334, boruta = False, opt = False)
    # estimator.fit(X_train, y_train)
    # print(mean_squared_error(estimator.predict(X_test), y_test, squared = False))



if __name__ == '__main__':
    from sklearn.datasets import load_boston
    import pandas as pd
    from yikit.feature_selection import BorutaPy
    from sklearn.ensemble import RandomForestRegressor

    boston = load_boston()
    X = pd.DataFrame(boston['data'], columns = boston['feature_names'])
    y = pd.Series(boston['target'], name = 'PRICE')
    print(BorutaPy(RandomForestRegressor(n_jobs=-1, random_state=334), random_state=334).fit(X, y).support_)


# PyPI登録流れ参考: https://qiita.com/kinpira/items/0a4e7c78fc5dd28bd695

"""
This is a setup.py script generated by py2applet
Usage:
    python setup.py py2app
"""

from setuptools import setup, find_packages
from shutil import copyfile
import os
import sys
import re

# 再帰回数に引っかかるのでとりあえず大きい数に．
sys.setrecursionlimit(10 ** 9)

# ------------------------ ここを変更 --------------------------------
PACKAGE_NAME = 'yikit'    # フォルダの名前も統一
DESCRIPTION = 'This is my own tool kit.'

# py2app用の変数
SRC = ['main.py']
DATA_FILES = ['LICENSE']
PKGS = []
ICON = os.path.join('icon', '{}.icns'.format(PACKAGE_NAME))
# --------------------------------------------------------------------

VERSION_PYTHON = '{0}.{1}'.format(sys.version_info.major, sys.version_info.minor)

# __init__.pyから読み込む
with open(os.path.join(PACKAGE_NAME, '__init__.py')) as f:
    init_text = f.read()
    VERSION = re.search(r'__version__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    LICENSE = re.search(r'__license__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    AUTHOR = re.search(r'__author__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    EMAIL = re.search(r'__author_email__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    ID = re.search(r'__user_id__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    APP_NAME = re.search(r'__app_name__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    url = re.search(r'__url__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)

assert VERSION
assert LICENSE
assert AUTHOR
assert EMAIL
assert ID
assert APP_NAME
assert url

'''
メモ
/.pyenv/versions/anaconda3-2019.03/envs/pymat/lib/python3.7/site-packages/PyQt5/uic/port_v2/ascii_upper.pyの28行目を書き換えた．
少なくともpython3.7においてstringオブジェクトはmaketrans関数を持っておらず，正しくはstr.maketransである．
python3.8で以前やったときはこのエラーがでなかった．
* 3.8にアップデートしてやってみる．
    → 3.8だとsetupのほうがエラー起きるので3.6でやってみる．
    3.6は起動後に関数が呼び出せないと言われる
setup.pyに関する参考サイト
* https://packaging.python.org/guides/distributing-packages-using-setuptools/#packages
py2app 0.23や0.22では動作確認
'''

if 'py2app' in sys.argv:
    alias = '-A' in sys.argv or '--alias' in sys.argv

    # 諸変数・定数の定義
    lib_path = os.path.join(os.environ['CONDA_PREFIX'], 'lib')
    fname_libpython = 'libpython{}.dylib'.format(VERSION_PYTHON)

    # libpython3.7.m.dylibだとエラーになるのであらかじめコピーしておく．
    path_original = os.path.join(lib_path, 'libpython{}m.dylib'.format(VERSION_PYTHON))
    path_converted = os.path.join(lib_path, fname_libpython)
    if os.path.exists(path_original) and not os.path.exists(path_converted):
        copyfile(path_original, path_converted)

    # 諸変数の準備
    dylib_files = [os.path.join(lib_path, f) for f in os.listdir(lib_path) if '.dylib' in f]
    contents_path = os.path.join('dist', '{}.app'.format(APP_NAME), 'Contents')
    frameworks_path = os.path.join(contents_path, 'Frameworks')

    OPTIONS = {
        'argv_emulation': False,
        'packages': PKGS,
        'iconfile': ICON,
        'plist':{
            'PyRuntimeLocations':[
                '@executable_path/../Frameworks/{}'.format(fname_libpython),
                os.path.join(lib_path, fname_libpython),
            ],
            'CFBundleName': APP_NAME,
            'CFBundleDisplayName': APP_NAME,
            'CFBundleGetInfoString': DESCRIPTION,
            'CFBundleIdentifier': "com.{0}.osx.{1}".format(ID, APP_NAME),
            'CFBundleVersion': VERSION,
            'CFBundleShortVersionString': VERSION,
            'NSHumanReadableCopyright': u"Copyright © 2021-, {}".format(AUTHOR)
        },
        # 'frameworks': dylib_files,
    }

    setup(
        name = APP_NAME,
        app = SRC,
        author = AUTHOR,
        author_email = EMAIL,
        version = VERSION,
        data_files = DATA_FILES,
        options = {'py2app': OPTIONS},
        setup_requires = ['py2app'],
        url = 'https://github.com/{0}/{1}'.format(ID, APP_NAME),
    )

    # aliasモードじゃないとき．
    # if not alias:
    #     {copyfile(f, os.path.join(frameworks_path, os.path.basename(f))) for f in dylib_files}
else:
    """
    参考: https://python-packaging-user-guide-ja.readthedocs.io/ja/latest/distributing.html#manifest-in
    """
    with open('requirements.txt') as requirements_file:
        install_requirements = requirements_file.read().splitlines()
    
    try:
        with open('README.md') as f:
            long_description = f.read()
    except IOError:
        long_description = ''

    setup(
        name = APP_NAME,
        version = VERSION,
        description = DESCRIPTION,
        long_description = long_description,
        long_description_content_type = 'text/markdown', # long_descriptionの形式を'text/plain', 'text/x-rst', 'text/markdown'のいずれかから指定．
        author = AUTHOR,
        author_email = EMAIL,
        maintainer = AUTHOR,
        maintainer_email = EMAIL,
        install_requires = install_requirements,
        url = url,
        keywords = '', # PyPIでの検索用キーワードをスペース区切りで指定．
        license = LICENSE,
        packages = find_packages(exclude=['example']),
        classifiers = [
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
        ],  # パッケージ(プロジェクト)の分類．https://pypi.org/classifiers/ に掲載されているものを指定可能．
    )


if __name__ == '__main__':
    from yikit.tools import is_notebook
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.datasets import load_boston
    from sklearn.model_selection import train_test_split
    import pandas as pd
    import matplotlib.pyplot as plt
    from pdb import set_trace
    import os

    # データセットの準備等
    boston = load_boston()
    X = pd.DataFrame(boston.data, columns = boston.feature_names)
    y = pd.Series(boston.target, name = 'PRICE')

    SEED = 334

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size = 0.2, random_state = SEED)


    # SummarizePI test
    # from yikit.tools.visualization import SummarizePI
    # from sklearn.inspection import permutation_importance
    
    # rf = RandomForestRegressor(random_state = SEED, n_jobs = -1)

    # rf.fit(X_train, y_train)

    # pi = permutation_importance(rf, X_test, y_test, random_state = SEED, n_jobs = -1)
    # spi = SummarizePI(pd.DataFrame(pi.importances, index = X.columns))
    # spi.get_figure()
    # plt.show()
    # plt.close()

    # get_dist_figure test
    from yikit.tools.visualization import get_dist_figure
    from ngboost import NGBRegressor
    ngb = NGBRegressor(random_state=SEED, verbose=False).fit(X_train, y_train)
    get_dist_figure(ngb.pred_dist(X_test), y_test, titles = ['a'] * len(y_test), verbose=False).savefig(os.path.join(os.path.dirname(__file__), 'sample_dist_figure.png'))

    # get_learning_curve_optuna
    # from yikit.tools.visualization import get_learning_curve_optuna
    # import optuna
    # from yikit.models import Objective
    # rf = RandomForestRegressor(random_state=SEED, n_jobs = -1)
    # objective = Objective(rf, X_train, y_train, random_state=SEED, n_jobs=-1)
    # study = optuna.create_study(sampler = objective.sampler, direction='maximize')
    # study.optimize(objective, n_trials=10)
    # get_learning_curve_optuna(study).savefig(os.path.join(os.path.dirname(__file__), 'sample_learning_curve_optuna.png'))


    # get_learning_curve_gb
    # from yikit.tools.visualization import get_learning_curve_gb
    # from ngboost import NGBRegressor
    # from sklearn.tree import DecisionTreeRegressor
    # ngb = NGBRegressor(random_state=SEED, Base=DecisionTreeRegressor(random_state=SEED))
    # ngb.fit(X_train, y_train, X_val=X_test, Y_val=y_test)
    # get_learning_curve_gb(ngb).savefig(os.path.join(os.path.dirname(__file__), 'sample_learning_curve_ngboost.png'))

    # from yikit.tools.visualization import get_learning_curve_gb
    # from lightgbm import LGBMRegressor
    # lgbm = LGBMRegressor(random_state=SEED, n_jobs=-1)
    # lgbm.fit(X_train, y_train, eval_set=[(X_train, y_train), (X_test, y_test)], eval_names=['train', 'test'])
    # get_learning_curve_gb(lgbm).savefig(os.path.join(os.path.dirname(__file__), 'sample_learning_curve_lightgbm.png'))
    
    
